--  Copyright (C) 1994-2001 Grady Booch and Simon Wright.
--  All Rights Reserved.
--
--      This program is free software; you can redistribute it
--      and/or modify it under the terms of the Ada Community
--      License which comes with this Library.
--
--      This program is distributed in the hope that it will be
--      useful, but WITHOUT ANY WARRANTY; without even the implied
--      warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
--      PURPOSE. See the Ada Community License for more details.
--      You should have received a copy of the Ada Community
--      License with this library, in the file named "Ada Community
--      License" or "ACL". If not, contact the author of this library
--      for a copy.
--

--  $RCSfile: bc-containers-maps-synchronized.adb,v $
--  $Revision: 1.11 $
--  $Date: 2001/09/12 05:59:38 $
--  $Author: simon $

with System;

package body BC.Containers.Maps.Synchronized is

   function "=" (L, R : Map) return Boolean is
      --  To avoid deadlock, make sure we always apply locks in the
      --  same order
      use type System.Address;
   begin
      if L'Address = R'Address then
         return True;
      elsif L'Address > R'Address then
         return R = L;
      else
         declare
            L_Lock : BC.Support.Synchronization.Read_Lock (L.The_Monitor);
            pragma Warnings (Off, L_Lock);
            R_Lock : BC.Support.Synchronization.Read_Lock (R.The_Monitor);
            pragma Warnings (Off, R_Lock);
         begin
            return Map_Base (L) = Map_Base (R);
         end;
      end if;
   end "=";

   procedure Bind (M : in out Map; K : Key; I : Item) is
      Lock : BC.Support.Synchronization.Write_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      Bind (Map_Base (M), K, I);
   end Bind;

   procedure Clear (M : in out Map) is
      Lock : BC.Support.Synchronization.Write_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      Clear (Map_Base (M));
   end Clear;

   function Available (M : Map) return Natural is
      Lock : BC.Support.Synchronization.Read_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Available (Map_Base (M));
   end Available;

   function Extent (M : Map) return Natural is
      Lock : BC.Support.Synchronization.Read_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Extent (Map_Base (M));
   end Extent;

   function Is_Bound (M : Map; K : Key) return Boolean is
      Lock : BC.Support.Synchronization.Read_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Is_Bound (Map_Base (M), K);
   end Is_Bound;

   function Is_Empty (M : Map) return Boolean is
      Lock : BC.Support.Synchronization.Read_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Is_Empty (Map_Base (M));
   end Is_Empty;

   function Item_Of (M : Map; K : Key) return Item is
      Lock : BC.Support.Synchronization.Read_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      return Item_Of (Map_Base (M), K);
   end Item_Of;

   procedure Lock (M : in out Map) is
   begin
      BC.Support.Synchronization.Seize_For_Reading (M.The_Monitor.all);
   end Lock;

   procedure Rebind
     (M : in out Map; K : Key; I : Item) is
      Lock : BC.Support.Synchronization.Write_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      Rebind (Map_Base (M), K, I);
   end Rebind;

   procedure Unbind (M : in out Map; K : Key) is
      Lock : BC.Support.Synchronization.Write_Lock (M.The_Monitor);
      pragma Warnings (Off, Lock);
   begin
      Unbind (Map_Base (M), K);
   end Unbind;

   procedure Unlock (M : in out Map) is
   begin
      BC.Support.Synchronization.Release_From_Reading (M.The_Monitor.all);
   end Unlock;

   procedure Initialize (M : in out Map) is
   begin
      Initialize (Map_Base (M));
      M.The_Monitor := new Monitor;
   end Initialize;

   procedure Adjust (M : in out Map) is
      --  Consider M := P;
      --  On entry to Adjust, M contains a bitwise copy of P, so that
      --  M.The_Monitor points to the same Monitor as P.The_Monitor.
   begin
      --  lock P's monitor
      Lock (M);
      --  make the deep copy
      Adjust (Map_Base (M));
      --  unlock P's monitor
      Unlock (M);
      --  create new monitor, semaphores for M
      M.The_Monitor := new Monitor;
   end Adjust;

   procedure Finalize (M : in out Map) is
   begin
      BC.Support.Synchronization.Delete (M.The_Monitor);
      Finalize (Map_Base (M));
   end Finalize;

   Empty_Container : Map;
   pragma Warnings (Off, Empty_Container);

   function Null_Container return Map is
   begin
      return Empty_Container;
   end Null_Container;

end BC.Containers.Maps.Synchronized;
